=== modified file 'NEWS'
--- NEWS	2006-05-23 07:51:49 +0000
+++ NEWS	2006-05-23 10:57:02 +0000
@@ -31,6 +31,10 @@
      ignored). 'bzr add -v' is unchanged and will report every ignored
      file. (Robert Collins).
 
+   * ftp now has a test server if medusa is installed. As part of testing,
+     ftp support has been improved, including support for supplying a
+     non-standard port. (John Arbash Meinel).
+
    * ``log --line`` shows the revision number.  (#5162, Alexander Belchenko)
 
   BUG FIXES:
@@ -39,6 +43,16 @@
       tests not to run.  (#32587, Erik Bågfors, Michael Ellerman, 
       Martin Pool)
 
+  INTERNALS:
+
+    * Combine the ignore rules into a single regex rather than looping over
+      them to reduce the threshold where  N^2 behaviour occurs in operations
+      like status. (Jan Hudec, Robert Collins).
+
+    * 'bzr push' should only push the ancestry of the current revision, not
+      all of the history in the repository. This is especially important for
+      shared repositories. (John Arbash Meinel)
+
 bzr 0.8.2  2006-05-17
   
   BUG FIXES:

=== modified file 'bzrlib/add.py'
--- bzrlib/add.py	2006-05-18 07:10:08 +0000
+++ bzrlib/add.py	2006-05-23 00:03:45 +0000
@@ -20,7 +20,6 @@
 from bzrlib.inventory import InventoryEntry
 from bzrlib.trace import mutter, note, warning
 from bzrlib.errors import NotBranchError
-from bzrlib.workingtree import is_control_file
 import bzrlib.osutils
 from bzrlib.workingtree import WorkingTree
 
@@ -52,24 +51,31 @@
     return file_list
 
 
-def add_action_null(inv, path, kind):
+def add_action_null(inv, parent_ie, path, kind):
     """Absorb add actions and do nothing."""
     pass
 
-def add_action_print(inv, path, kind):
+
+def add_action_print(inv, parent_ie, path, kind):
     """Print a line to stdout for each file that would be added."""
     print "added", bzrlib.osutils.quotefn(path)
 
-def add_action_add(inv, path, kind):
+
+def add_action_add(inv, parent_ie, path, kind):
     """Add each file to the given inventory. Produce no output."""
-    entry = inv.add_path(path, kind=kind)
-    mutter("added %r kind %r file_id={%s}" % (path, kind, entry.file_id))
-
-
-def add_action_add_and_print(inv, path, kind):
+    if parent_ie is not None:
+        entry = bzrlib.inventory.make_entry(
+            kind, bzrlib.osutils.basename(path),  parent_ie.file_id)
+        inv.add(entry)
+    else:
+        entry = inv.add_path(path, kind=kind)
+    # mutter("added %r kind %r file_id={%s}" % (path, kind, entry.file_id))
+
+
+def add_action_add_and_print(inv, parent_ie, path, kind):
     """Add each file to the given inventory, and print a line to stdout."""
-    add_action_add(inv, path, kind)
-    add_action_print(inv, path, kind)
+    add_action_add(inv, parent_ie, path, kind)
+    add_action_print(inv, parent_ie, path, kind)
 
 
 def smart_add(file_list, recurse=True, action=add_action_add):
@@ -98,41 +104,65 @@
     import os, errno
     from bzrlib.errors import BadFileKindError, ForbiddenFileError
     assert isinstance(recurse, bool)
-
-    file_list = _prepare_file_list(file_list)
-    user_list = file_list[:]
+    
+    prepared_list = _prepare_file_list(file_list)
+    mutter("smart add of %r, originally %r", prepared_list, file_list)
     inv = tree.read_working_inventory()
     added = []
     ignored = {}
-
-    for f in file_list:
-        rf = tree.relpath(f)
-        af = tree.abspath(rf)
-
+    user_files = set()
+    files_to_add = []
+
+    # validate user file paths and convert all paths to tree 
+    # relative : its cheaper to make a tree relative path an abspath
+    # than to convert an abspath to tree relative.
+    for filepath in prepared_list:
+        rf = tree.relpath(filepath)
+        user_files.add(rf)
+        files_to_add.append((rf, None))
+        # validate user parameters. Our recursive code avoids adding new files
+        # that need such validation 
+        if tree.is_control_filename(rf):
+            raise ForbiddenFileError('cannot add control file %s' % filepath)
+
+    for filepath, parent_ie in files_to_add:
+        # filepath is tree-relative
+        abspath = tree.abspath(filepath)
+
+        # find the kind of the path being added. This is not
+        # currently determined when we list directories 
+        # recursively, but in theory we can determine while 
+        # doing the directory listing on *some* platformans.
+        # TODO: a safe, portable, clean interface which will 
+        # be faster than os.listdir() + stat. Specifically,
+        # readdir - dirent.d_type supplies the file type when
+        # it is defined. (Apparently Mac OSX has the field but
+        # does not fill it in ?!) Robert C, Martin P.
         try:
-            kind = bzrlib.osutils.file_kind(af)
+            kind = bzrlib.osutils.file_kind(abspath)
         except OSError, e:
             if hasattr(e, 'errno') and e.errno == errno.ENOENT:
-                raise errors.NoSuchFile(rf)
+                raise errors.NoSuchFile(abspath)
             raise
 
+        # we need to call this to determine the inventory kind to create.
         if not InventoryEntry.versionable_kind(kind):
-            if f in user_list:
-                raise BadFileKindError("cannot add %s of type %s" % (f, kind))
+            if filepath in user_files:
+                raise BadFileKindError("cannot add %s of type %s" % (abspath, kind))
             else:
-                warning("skipping %s (can't add file of kind '%s')", f, kind)
+                warning("skipping %s (can't add file of kind '%s')", abspath, kind)
                 continue
 
-        mutter("smart add of %r, abs=%r", f, af)
-        
-        if tree.is_control_filename(af):
-            raise ForbiddenFileError('cannot add control file %s' % f)
-            
-        versioned = (inv.path2id(rf) != None)
+        if parent_ie is not None:
+            versioned = bzrlib.osutils.basename(filepath) in parent_ie.children
+        else:
+            # without the parent ie, use the relatively slower inventory 
+            # probing method
+            versioned = inv.has_filename(filepath)
 
         if kind == 'directory':
             try:
-                sub_branch = WorkingTree.open(af)
+                sub_branch = bzrlib.bzrdir.BzrDir.open(abspath)
                 sub_tree = True
             except NotBranchError:
                 sub_tree = False
@@ -141,51 +171,92 @@
         else:
             sub_tree = False
 
-        if rf == '':
-            mutter("tree root doesn't need to be added")
+        if filepath == '':
+            # mutter("tree root doesn't need to be added")
             sub_tree = False
         elif versioned:
-            mutter("%r is already versioned", f)
+            pass
+            # mutter("%r is already versioned", abspath)
         elif sub_tree:
-            mutter("%r is a bzr tree", f)
+            mutter("%r is a nested bzr tree", abspath)
         else:
-            added.extend(__add_one(tree, inv, rf, kind, action))
+            added.extend(__add_one(tree, inv, parent_ie, filepath, kind, action))
 
         if kind == 'directory' and recurse and not sub_tree:
-            for subf in os.listdir(af):
-                subp = bzrlib.osutils.pathjoin(rf, subf)
+            try:
+                if parent_ie is not None:
+                    # must be present:
+                    this_ie = parent_ie.children[bzrlib.osutils.basename(filepath)]
+                else:
+                    # without the parent ie, use the relatively slower inventory 
+                    # probing method
+                    this_id = inv.path2id(filepath)
+                    if this_id is None:
+                        this_ie = None
+                    else:
+                        this_ie = inv[this_id]
+            except KeyError:
+                this_ie = None
+
+            for subf in os.listdir(abspath):
+                # here we could use TreeDirectory rather than 
+                # string concatenation.
+                subp = bzrlib.osutils.pathjoin(filepath, subf)
+                # TODO: is_control_filename is very slow. Make it faster. 
+                # TreeDirectory.is_control_filename could also make this 
+                # faster - its impossible for a non root dir to have a 
+                # control file.
                 if tree.is_control_filename(subp):
                     mutter("skip control directory %r", subp)
                 else:
+                    # ignore while selecting files - if we globbed in the
+                    # outer loop we would ignore user files.
                     ignore_glob = tree.is_ignored(subp)
                     if ignore_glob is not None:
-                        mutter("skip ignored sub-file %r", subp)
+                        # mutter("skip ignored sub-file %r", subp)
                         if ignore_glob not in ignored:
                             ignored[ignore_glob] = []
                         ignored[ignore_glob].append(subp)
                     else:
-                        mutter("queue to add sub-file %r", subp)
-                        file_list.append(tree.abspath(subp))
+                        #mutter("queue to add sub-file %r", subp)
+                        files_to_add.append((subp, this_ie))
 
-    mutter('added %d entries', len(added))
-    
     if len(added) > 0:
         tree._write_inventory(inv)
-
     return added, ignored
 
 
-def __add_one(tree, inv, path, kind, action):
-    """Add a file or directory, automatically add unversioned parents."""
+def __add_one(tree, inv, parent_ie, path, kind, action):
+    """Add a new entry to the inventory and automatically add unversioned parents.
+
+    Actual adding of the entry is delegated to the action callback.
+
+    :param inv: Inventory which will receive the new entry.
+    :param parent_ie: Parent inventory entry if known, or None.  If
+    None, the parent is looked up by name and used if present, otherwise
+    it is recursively added.
+    :param kind: Kind of new entry (file, directory, etc)
+    :param action: callback(inv, parent_ie, path, kind); return ignored.
+    :returns: A list of paths which have been added.
+    """
 
     # Nothing to do if path is already versioned.
     # This is safe from infinite recursion because the tree root is
     # always versioned.
-    if inv.path2id(path) != None:
-        return []
-
-    # add parent
-    added = __add_one(tree, inv, dirname(path), 'directory', action)
-    action(inv, path, kind)
+    if parent_ie is not None:
+        # we have a parent ie already
+        added = []
+    else:
+        # slower but does not need parent_ie
+        if inv.has_filename(path):
+            return []
+        # add parent
+        added = __add_one(tree, inv, None, dirname(path), 'directory', action)
+        parent_id = inv.path2id(dirname(path))
+        if parent_id is not None:
+            parent_ie = inv[inv.path2id(dirname(path))]
+        else:
+            parent_ie = None
+    action(inv, parent_ie, path, kind)
 
     return added + [path]

=== modified file 'bzrlib/benchmarks/__init__.py'
--- bzrlib/benchmarks/__init__.py	2006-05-16 06:45:43 +0000
+++ bzrlib/benchmarks/__init__.py	2006-05-22 09:17:55 +0000
@@ -18,11 +18,35 @@
 """Benchmark test suite for bzr."""
 
 from bzrlib.tests import TestLoader
+from bzrlib.tests.blackbox import ExternalBase
+
+class Benchmark(ExternalBase):
+
+    def make_kernel_like_tree(self):
+        """Setup a temporary tree roughly like a kernel tree."""
+        # a kernel tree has ~10000 and 500 directory, with most files around 
+        # 3-4 levels deep. 
+        # we simulate this by three levels of dirs named 0-7, givin 512 dirs,
+        # and 20 files each.
+        self.run_bzr('init')
+        files = []
+        for outer in range(8):
+            files.append("%s/" % outer)
+            for middle in range(8):
+                files.append("%s/%s/" % (outer, middle))
+                for inner in range(8):
+                    prefix = "%s/%s/%s/" % (outer, middle, inner)
+                    files.append(prefix)
+                    files.extend([prefix + str(foo) for foo in range(20)])
+        self.build_tree(files)
 
 
 def test_suite():
     """Build and return a TestSuite which contains benchmark tests only."""
     testmod_names = [ \
                    'bzrlib.benchmarks.bench_add',
+                   'bzrlib.benchmarks.bench_checkout',
+                   'bzrlib.benchmarks.bench_commit',
+                   'bzrlib.benchmarks.bench_status',
                    ]
     return TestLoader().loadTestsFromModuleNames(testmod_names)

=== modified file 'bzrlib/benchmarks/bench_add.py'
--- bzrlib/benchmarks/bench_add.py	2006-05-17 03:40:51 +0000
+++ bzrlib/benchmarks/bench_add.py	2006-05-23 00:07:45 +0000
@@ -16,28 +16,14 @@
 """Tests for bzr add performance."""
 
 
-from bzrlib.tests.blackbox import ExternalBase
-
-
-class TestAdd(ExternalBase):
+from bzrlib.benchmarks import Benchmark
+
+
+class AddBenchmark(Benchmark):
 
     def test_one_add_kernel_like_tree(self):
         """Adding a kernel sized tree should be bearable (<5secs) fast.""" 
-        # a kernel tree has ~10000 and 500 directory, with most files around 
-        # 3-4 levels deep. 
-        # we simulate this by three levels of dirs named 0-7, givin 512 dirs,
-        # and 20 files each.
-        # on roberts machine this originally took 25936ms/32244ms
-        # after quick tuning (more needed):       14640ms/19677ms 
-        self.run_bzr('init')
-        files = []
-        for outer in range(8):
-            files.append("%s/" % outer)
-            for middle in range(8):
-                files.append("%s/%s/" % (outer, middle))
-                for inner in range(8):
-                    prefix = "%s/%s/%s/" % (outer, middle, inner)
-                    files.append(prefix)
-                    files.extend([prefix + str(foo) for foo in range(20)])
-        self.build_tree(files)
+        self.make_kernel_like_tree()
+        # on roberts machine this originally took:  25936ms/32244ms
+        # after making smart_add use the parent_ie:  5033ms/ 9368ms
         self.time(self.run_bzr, 'add')

=== added file 'bzrlib/benchmarks/bench_checkout.py'
--- bzrlib/benchmarks/bench_checkout.py	1970-01-01 00:00:00 +0000
+++ bzrlib/benchmarks/bench_checkout.py	2006-05-22 09:17:55 +0000
@@ -0,0 +1,30 @@
+# Copyright (C) 2006 by Canonical Ltd
+#
+# This program is free software; you can redistribute it and/or modify
+# it under the terms of the GNU General Public License version 2 as published by
+# the Free Software Foundation.
+#
+# This program is distributed in the hope that it will be useful,
+# but WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+# GNU General Public License for more details.
+#
+# You should have received a copy of the GNU General Public License
+# along with this program; if not, write to the Free Software
+# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+
+"""Tests for bzr tree building (checkout) performance."""
+
+
+from bzrlib.benchmarks import Benchmark
+
+
+class CheckoutBenchmark(Benchmark):
+
+    def test_build_kernel_like_tree(self):
+        """Checkout of a clean kernel sized tree should be (<10secs)."""
+        self.make_kernel_tree()
+        self.run_bzr('add')
+        self.run_bzr('commit', '-m', 'first post')
+        # on robertc's machine the first sample of this took 105079ms/205417ms
+        self.time(self.run_bzr, 'checkout', '--lightweight', '.', 'acheckout')

=== added file 'bzrlib/benchmarks/bench_commit.py'
--- bzrlib/benchmarks/bench_commit.py	1970-01-01 00:00:00 +0000
+++ bzrlib/benchmarks/bench_commit.py	2006-05-22 09:17:55 +0000
@@ -0,0 +1,29 @@
+# Copyright (C) 2006 by Canonical Ltd
+#
+# This program is free software; you can redistribute it and/or modify
+# it under the terms of the GNU General Public License version 2 as published by
+# the Free Software Foundation.
+#
+# This program is distributed in the hope that it will be useful,
+# but WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+# GNU General Public License for more details.
+#
+# You should have received a copy of the GNU General Public License
+# along with this program; if not, write to the Free Software
+# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+
+"""Benchmarks of bzr commit."""
+
+
+from bzrlib.benchmarks import Benchmark
+
+
+class CommitBenchmark(Benchmark):
+
+    def test_commit_kernel_like_tree(self):
+        """Commit of a fresh import of a clean kernel sized tree."""
+        self.make_kernel_like_tree()
+        self.run_bzr('add')
+        # on robertc's machine the first sample of this took 59750ms/77682ms
+        self.time(self.run_bzr, 'commit', '-m', 'first post')

=== added file 'bzrlib/benchmarks/bench_status.py'
--- bzrlib/benchmarks/bench_status.py	1970-01-01 00:00:00 +0000
+++ bzrlib/benchmarks/bench_status.py	2006-05-22 09:17:55 +0000
@@ -0,0 +1,31 @@
+# Copyright (C) 2006 by Canonical Ltd
+#
+# This program is free software; you can redistribute it and/or modify
+# it under the terms of the GNU General Public License version 2 as published by
+# the Free Software Foundation.
+#
+# This program is distributed in the hope that it will be useful,
+# but WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+# GNU General Public License for more details.
+#
+# You should have received a copy of the GNU General Public License
+# along with this program; if not, write to the Free Software
+# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+
+"""Tests for bzr status performance."""
+
+
+from bzrlib.benchmarks import Benchmark
+
+
+class StatusBenchmark(Benchmark):
+
+    def test_no_ignored_unknown_kernel_like_tree(self):
+        """Status in a kernel sized tree with no ignored or unknowns.
+        
+        This should be bearable (<2secs) fast.""" 
+        self.make_kernel_like_tree()
+        self.run_bzr('add')
+        # on robertc's machine the first sample of this took 1687ms/15994ms
+        self.time(self.run_bzr, 'status')

=== modified file 'bzrlib/builtins.py'
--- bzrlib/builtins.py	2006-05-19 06:04:00 +0000
+++ bzrlib/builtins.py	2006-05-21 17:56:17 +0000
@@ -517,7 +517,8 @@
                         if new_transport.base == transport.base:
                             raise BzrCommandError("Could not create "
                                                   "path prefix.")
-            dir_to = br_from.bzrdir.clone(location)
+            dir_to = br_from.bzrdir.clone(location,
+                revision_id=br_from.last_revision())
             br_to = dir_to.open_branch()
             count = len(br_to.revision_history())
         else:
@@ -1948,12 +1949,19 @@
     
     merge refuses to run if there are any uncommitted changes, unless
     --force is given.
+
+    The following merge types are available:
     """
     takes_args = ['branch?']
     takes_options = ['revision', 'force', 'merge-type', 'reprocess', 'remember',
                      Option('show-base', help="Show base revision text in "
                             "conflicts")]
 
+    def help(self):
+        from merge import merge_type_help
+        from inspect import getdoc
+        return getdoc(self) + '\n' + merge_type_help() 
+
     def run(self, branch=None, revision=None, force=False, merge_type=None,
             show_base=False, reprocess=False, remember=False):
         if merge_type is None:
@@ -2020,12 +2028,35 @@
 
 class cmd_remerge(Command):
     """Redo a merge.
-    """
+
+    Use this if you want to try a different merge technique while resolving
+    conflicts.  Some merge techniques are better than others, and remerge 
+    lets you try different ones on different files.
+
+    The options for remerge have the same meaning and defaults as the ones for
+    merge.  The difference is that remerge can (only) be run when there is a
+    pending merge, and it lets you specify particular files.
+
+    Examples:
+    $ bzr remerge --show-base
+        Re-do the merge of all conflicted files, and show the base text in
+        conflict regions, in addition to the usual THIS and OTHER texts.
+
+    $ bzr remerge --merge-type weave --reprocess foobar
+        Re-do the merge of "foobar", using the weave merge algorithm, with
+        additional processing to reduce the size of conflict regions.
+    
+    The following merge types are available:"""
     takes_args = ['file*']
     takes_options = ['merge-type', 'reprocess',
                      Option('show-base', help="Show base revision text in "
                             "conflicts")]
 
+    def help(self):
+        from merge import merge_type_help
+        from inspect import getdoc
+        return getdoc(self) + '\n' + merge_type_help() 
+
     def run(self, file_list=None, merge_type=None, show_base=False,
             reprocess=False):
         from bzrlib.merge import merge_inner, transform_tree

=== modified file 'bzrlib/bzrdir.py'
--- bzrlib/bzrdir.py	2006-05-17 04:09:58 +0000
+++ bzrlib/bzrdir.py	2006-05-19 07:57:09 +0000
@@ -412,6 +412,25 @@
         self.transport = _transport.clone('.bzr')
         self.root_transport = _transport
 
+    def is_control_filename(self, filename):
+        """True if filename is the name of a path which is reserved for bzrdir's.
+        
+        :param filename: A filename within the root transport of this bzrdir.
+
+        This is true IF and ONLY IF the filename is part of the namespace reserved
+        for bzr control dirs. Currently this is the '.bzr' directory in the root
+        of the root_transport. it is expected that plugins will need to extend
+        this in the future - for instance to make bzr talk with svn working
+        trees.
+        """
+        # this might be better on the BzrDirFormat class because it refers to 
+        # all the possible bzrdir disk formats. 
+        # This method is tested via the workingtree is_control_filename tests- 
+        # it was extractd from WorkingTree.is_control_filename. If the methods
+        # contract is extended beyond the current trivial  implementation please
+        # add new tests for it to the appropriate place.
+        return filename == '.bzr' or filename.startswith('.bzr/')
+
     def needs_format_conversion(self, format=None):
         """Return true if this bzrdir needs convert_format run on it.
         

=== modified file 'bzrlib/errors.py'
--- bzrlib/errors.py	2006-05-15 22:54:09 +0000
+++ bzrlib/errors.py	2006-05-20 17:48:41 +0000
@@ -34,7 +34,7 @@
 ... except:
 ...   print sys.exc_type
 ...   print sys.exc_value
-...   path = getattr(sys.exc_value, 'path')
+...   path = getattr(sys.exc_value, 'path', None)
 ...   if path is not None:
 ...     print path
 bzrlib.errors.NotBranchError

=== modified file 'bzrlib/inventory.py'
--- bzrlib/inventory.py	2006-05-12 08:29:23 +0000
+++ bzrlib/inventory.py	2006-05-19 13:53:22 +0000
@@ -993,20 +993,18 @@
         return entry
 
 
-    def add_path(self, relpath, kind, file_id=None):
+    def add_path(self, relpath, kind, file_id=None, parent_id=None):
         """Add entry from a path.
 
         The immediate parent must already be versioned.
 
         Returns the new entry object."""
-        from bzrlib.workingtree import gen_file_id
         
         parts = bzrlib.osutils.splitpath(relpath)
 
-        if file_id == None:
-            file_id = gen_file_id(relpath)
-
         if len(parts) == 0:
+            if file_id is None:
+                file_id = bzrlib.workingtree.gen_root_id()
             self.root = RootEntry(file_id)
             self._byid = {self.root.file_id: self.root}
             return
@@ -1015,17 +1013,9 @@
             parent_id = self.path2id(parent_path)
             if parent_id == None:
                 raise NotVersionedError(path=parent_path)
-        if kind == 'directory':
-            ie = InventoryDirectory(file_id, parts[-1], parent_id)
-        elif kind == 'file':
-            ie = InventoryFile(file_id, parts[-1], parent_id)
-        elif kind == 'symlink':
-            ie = InventoryLink(file_id, parts[-1], parent_id)
-        else:
-            raise BzrError("unknown kind %r" % kind)
+        ie = make_entry(kind, parts[-1], parent_id, file_id)
         return self.add(ie)
 
-
     def __delitem__(self, file_id):
         """Remove entry by id.
 
@@ -1127,12 +1117,12 @@
         This returns the entry of the last component in the path,
         which may be either a file or a directory.
 
-        Returns None iff the path is not found.
+        Returns None IFF the path is not found.
         """
         if isinstance(name, types.StringTypes):
             name = splitpath(name)
 
-        mutter("lookup path %r" % name)
+        # mutter("lookup path %r" % name)
 
         parent = self.root
         for f in name:
@@ -1186,6 +1176,25 @@
         file_ie.parent_id = new_parent_id
 
 
+def make_entry(kind, name, parent_id, file_id=None):
+    """Create an inventory entry.
+
+    :param kind: the type of inventory entry to create.
+    :param name: the basename of the entry.
+    :param parent_id: the parent_id of the entry.
+    :param file_id: the file_id to use. if None, one will be created.
+    """
+    if file_id is None:
+        file_id = bzrlib.workingtree.gen_file_id(name)
+    if kind == 'directory':
+        return InventoryDirectory(file_id, name, parent_id)
+    elif kind == 'file':
+        return InventoryFile(file_id, name, parent_id)
+    elif kind == 'symlink':
+        return InventoryLink(file_id, name, parent_id)
+    else:
+        raise BzrError("unknown kind %r" % kind)
+
 
 
 _NAME_RE = None

=== modified file 'bzrlib/merge.py'
--- bzrlib/merge.py	2006-05-05 01:15:53 +0000
+++ bzrlib/merge.py	2006-05-19 04:00:05 +0000
@@ -919,3 +919,9 @@
                      "diff3": (Diff3Merger,  "Merge using external diff3"),
                      'weave': (WeaveMerger, "Weave-based merge")
               }
+
+
+def merge_type_help():
+    templ = '%s%%7s: %%s' % (' '*12)
+    lines = [templ % (f[0], f[1][1]) for f in merge_types.iteritems()]
+    return '\n'.join(lines)

=== modified file 'bzrlib/option.py'
--- bzrlib/option.py	2006-05-04 06:03:00 +0000
+++ bzrlib/option.py	2006-05-19 04:00:05 +0000
@@ -196,7 +196,8 @@
 _global_option('line', help='Use log format with one line per revision. Same as --log-format line')
 _global_option('root', type=str)
 _global_option('no-backup')
-_global_option('merge-type', type=_parse_merge_type)
+_global_option('merge-type', type=_parse_merge_type, 
+               help='Select a particular merge algorithm')
 _global_option('pattern', type=str)
 _global_option('quiet')
 _global_option('remember', help='Remember the specified location as a'

=== modified file 'bzrlib/tests/blackbox/__init__.py'
--- bzrlib/tests/blackbox/__init__.py	2006-05-18 08:21:57 +0000
+++ bzrlib/tests/blackbox/__init__.py	2006-05-18 23:48:54 +0000
@@ -50,6 +50,7 @@
                      'bzrlib.tests.blackbox.test_export',
                      'bzrlib.tests.blackbox.test_find_merge_base',
                      'bzrlib.tests.blackbox.test_help',
+                     'bzrlib.tests.blackbox.test_ignored',
                      'bzrlib.tests.blackbox.test_info',
                      'bzrlib.tests.blackbox.test_init',
                      'bzrlib.tests.blackbox.test_log',

=== added file 'bzrlib/tests/blackbox/test_ignored.py'
--- bzrlib/tests/blackbox/test_ignored.py	1970-01-01 00:00:00 +0000
+++ bzrlib/tests/blackbox/test_ignored.py	2006-05-18 23:48:54 +0000
@@ -0,0 +1,40 @@
+# Copyright (C) 2006 by Canonical Ltd
+#
+# This program is free software; you can redistribute it and/or modify
+# it under the terms of the GNU General Public License as published by
+# the Free Software Foundation; either version 2 of the License, or
+# (at your option) any later version.
+#
+# This program is distributed in the hope that it will be useful,
+# but WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+# GNU General Public License for more details.
+#
+# You should have received a copy of the GNU General Public License
+# along with this program; if not, write to the Free Software
+# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+#
+
+"""Tests of the 'bzr ignored' command."""
+
+from bzrlib.tests.blackbox import ExternalBase
+
+
+class TestIgnored(ExternalBase):
+        
+    def test_ignored_added_file(self):
+        """'bzr ignored' should not list versioned files."""
+        # this test can go in favour of a more general ui test at some point
+        # as it is actually testing the internals layer and should not be.
+        # There are no other 'ignored' tests though, so it should be retained
+        # until some are written.
+        tree = self.make_branch_and_tree('.')
+        self.build_tree(['foo.pyc'])
+        # ensure that foo.pyc is ignored
+        self.build_tree_contents([('.bzrignore', 'foo.pyc')])
+        self.assertTrue(tree.is_ignored('foo.pyc'))
+        # now add it and check the ui does not show it.
+        tree.add('foo.pyc')
+        out, err = self.run_bzr('ignored')
+        self.assertEqual('', out)
+        self.assertEqual('', err)

=== modified file 'bzrlib/tests/blackbox/test_push.py'
--- bzrlib/tests/blackbox/test_push.py	2006-05-15 05:06:57 +0000
+++ bzrlib/tests/blackbox/test_push.py	2006-05-21 13:58:58 +0000
@@ -22,9 +22,12 @@
 
 import bzrlib
 from bzrlib.branch import Branch
+from bzrlib.bzrdir import BzrDirMetaFormat1
 from bzrlib.osutils import abspath
+from bzrlib.repository import RepositoryFormatKnit1
 from bzrlib.tests.blackbox import ExternalBase
 from bzrlib.uncommit import uncommit
+from bzrlib.workingtree import WorkingTree
 
 
 class TestPush(ExternalBase):
@@ -94,3 +97,43 @@
         os.chdir('..')
         self.assertEqual('', out)
         self.assertEqual('1 revision(s) pushed.\n', err)
+
+    def test_push_only_pushes_history(self):
+        # Knit branches should only push the history for the current revision.
+        format = BzrDirMetaFormat1()
+        format.repository_format = RepositoryFormatKnit1()
+        shared_repo = self.make_repository('repo', format=format, shared=True)
+        shared_repo.set_make_working_trees(True)
+
+        def make_shared_tree(path):
+            shared_repo.bzrdir.root_transport.mkdir(path)
+            shared_repo.bzrdir.create_branch_convenience('repo/' + path)
+            return WorkingTree.open('repo/' + path)
+        tree_a = make_shared_tree('a')
+        self.build_tree(['repo/a/file'])
+        tree_a.add('file')
+        tree_a.commit('commit a-1', rev_id='a-1')
+        f = open('repo/a/file', 'ab')
+        f.write('more stuff\n')
+        f.close()
+        tree_a.commit('commit a-2', rev_id='a-2')
+
+        tree_b = make_shared_tree('b')
+        self.build_tree(['repo/b/file'])
+        tree_b.add('file')
+        tree_b.commit('commit b-1', rev_id='b-1')
+
+        self.assertTrue(shared_repo.has_revision('a-1'))
+        self.assertTrue(shared_repo.has_revision('a-2'))
+        self.assertTrue(shared_repo.has_revision('b-1'))
+
+        # Now that we have a repository with shared files, make sure
+        # that things aren't copied out by a 'push'
+        os.chdir('repo/b')
+        self.run_bzr('push', '../../push-b')
+        pushed_tree = WorkingTree.open('../../push-b')
+        pushed_repo = pushed_tree.branch.repository
+        self.assertFalse(pushed_repo.has_revision('a-1'))
+        self.assertFalse(pushed_repo.has_revision('a-2'))
+        self.assertTrue(pushed_repo.has_revision('b-1'))
+

=== modified file 'bzrlib/tests/test_inv.py'
--- bzrlib/tests/test_inv.py	2006-05-12 07:36:45 +0000
+++ bzrlib/tests/test_inv.py	2006-05-19 13:53:22 +0000
@@ -61,7 +61,6 @@
         
         self.assert_('src-id' in inv)
 
-
     def test_version(self):
         """Inventory remembers the text's version."""
         inv = Inventory()
@@ -134,6 +133,13 @@
         link = inventory.InventoryLink('123', 'hello.c', ROOT_ID)
         self.failIf(link.has_text())
 
+    def test_make_entry(self):
+        self.assertIsInstance(inventory.make_entry("file", "name", ROOT_ID),
+            inventory.InventoryFile)
+        self.assertIsInstance(inventory.make_entry("symlink", "name", ROOT_ID),
+            inventory.InventoryLink)
+        self.assertIsInstance(inventory.make_entry("directory", "name", ROOT_ID),
+            inventory.InventoryDirectory)
 
 class TestEntryDiffing(TestCaseWithTransport):
 

=== modified file 'bzrlib/tests/test_smart_add.py'
--- bzrlib/tests/test_smart_add.py	2006-02-06 10:31:46 +0000
+++ bzrlib/tests/test_smart_add.py	2006-05-19 13:53:22 +0000
@@ -208,7 +208,7 @@
         inv = Inventory()
         stdout = StringIO()
 
-        self.apply_redirected(None, stdout, None, action, inv, 'path', 'file')
+        self.apply_redirected(None, stdout, None, action, inv, None, 'path', 'file')
         self.assertEqual(stdout.getvalue(), output)
 
         if should_add:

=== modified file 'bzrlib/tests/test_transport_implementations.py'
--- bzrlib/tests/test_transport_implementations.py	2006-04-20 02:42:17 +0000
+++ bzrlib/tests/test_transport_implementations.py	2006-05-16 20:14:27 +0000
@@ -704,9 +704,10 @@
         except (ConnectionError, NoSuchFile), e:
             pass
         except (Exception), e:
-            self.failIf(True, 'Wrong exception thrown: %s' % e)
+            self.fail('Wrong exception thrown (%s): %s' 
+                        % (e.__class__.__name__, e))
         else:
-            self.failIf(True, 'Did not get the expected exception.')
+            self.fail('Did not get the expected ConnectionError or NoSuchFile.')
 
     def test_stat(self):
         # TODO: Test stat, just try once, and if it throws, stop testing

=== modified file 'bzrlib/tests/test_workingtree.py'
--- bzrlib/tests/test_workingtree.py	2006-04-18 22:41:16 +0000
+++ bzrlib/tests/test_workingtree.py	2006-05-19 06:00:00 +0000
@@ -234,3 +234,83 @@
         os.mkdir('lala.OTHER')
         expected = ContentsConflict('lala', file_id='lala-id')
         self.assertEqual(list(tree.conflicts()), [expected])
+
+
+class TestNonFormatSpecificCode(TestCaseWithTransport):
+    """This class contains tests of workingtree that are not format specific."""
+
+    
+    def test_gen_file_id(self):
+        self.assertStartsWith(bzrlib.workingtree.gen_file_id('bar'), 'bar-')
+        self.assertStartsWith(bzrlib.workingtree.gen_file_id('Mwoo oof\t m'), 'Mwoooofm-')
+        self.assertStartsWith(bzrlib.workingtree.gen_file_id('..gam.py'), 'gam.py-')
+        self.assertStartsWith(bzrlib.workingtree.gen_file_id('..Mwoo oof\t m'), 'Mwoooofm-')
+
+    def test_next_id_suffix(self):
+        bzrlib.workingtree._gen_id_suffix = None
+        bzrlib.workingtree._next_id_suffix()
+        self.assertNotEqual(None, bzrlib.workingtree._gen_id_suffix)
+        bzrlib.workingtree._gen_id_suffix = "foo-"
+        bzrlib.workingtree._gen_id_serial = 1
+        self.assertEqual("foo-2", bzrlib.workingtree._next_id_suffix())
+        self.assertEqual("foo-3", bzrlib.workingtree._next_id_suffix())
+        self.assertEqual("foo-4", bzrlib.workingtree._next_id_suffix())
+        self.assertEqual("foo-5", bzrlib.workingtree._next_id_suffix())
+        self.assertEqual("foo-6", bzrlib.workingtree._next_id_suffix())
+        self.assertEqual("foo-7", bzrlib.workingtree._next_id_suffix())
+        self.assertEqual("foo-8", bzrlib.workingtree._next_id_suffix())
+        self.assertEqual("foo-9", bzrlib.workingtree._next_id_suffix())
+        self.assertEqual("foo-10", bzrlib.workingtree._next_id_suffix())
+
+    def test__translate_ignore_rule(self):
+        tree = self.make_branch_and_tree('.')
+        # translation should return the regex, the number of groups in it,
+        # and the original rule in a tuple.
+        # there are three sorts of ignore rules:
+        # root only - regex is the rule itself without the leading ./
+        self.assertEqual(
+            "(rootdirrule$)", 
+            tree._translate_ignore_rule("./rootdirrule"))
+        # full path - regex is the rule itself
+        self.assertEqual(
+            "(path\\/to\\/file$)",
+            tree._translate_ignore_rule("path/to/file"))
+        # basename only rule - regex is a rule that ignores everything up
+        # to the last / in the filename
+        self.assertEqual(
+            "((?:.*/)?(?!.*/)basenamerule$)",
+            tree._translate_ignore_rule("basenamerule"))
+
+    def test__combine_ignore_rules(self):
+        tree = self.make_branch_and_tree('.')
+        # the combined ignore regexs need the outer group indices
+        # placed in a dictionary with the rules that were combined.
+        # an empty set of rules
+        # this is returned as a list of combined regex,rule sets, because
+        # python has a limit of 100 combined regexes.
+        compiled_rules = tree._combine_ignore_rules([])
+        self.assertEqual([], compiled_rules)
+        # one of each type of rule.
+        compiled_rules = tree._combine_ignore_rules(
+            ["rule1", "rule/two", "./three"])[0]
+        # what type *is* the compiled regex to do an isinstance of ?
+        self.assertEqual(3, compiled_rules[0].groups)
+        self.assertEqual(
+            {0:"rule1",1:"rule/two",2:"./three"},
+            compiled_rules[1])
+
+    def test__combine_ignore_rules_grouping(self):
+        tree = self.make_branch_and_tree('.')
+        # when there are too many rules, the output is split into groups of 100
+        rules = []
+        for index in range(198):
+            rules.append('foo')
+        self.assertEqual(2, len(tree._combine_ignore_rules(rules)))
+
+    def test__get_ignore_rules_as_regex(self):
+        tree = self.make_branch_and_tree('.')
+        # test against the default rules.
+        reference_output = tree._combine_ignore_rules(bzrlib.DEFAULT_IGNORE)[0]
+        regex_rules = tree._get_ignore_rules_as_regex()[0]
+        self.assertEqual(len(reference_output[1]), regex_rules[0].groups)
+        self.assertEqual(reference_output[1], regex_rules[1])

=== modified file 'bzrlib/tests/workingtree_implementations/__init__.py'
--- bzrlib/tests/workingtree_implementations/__init__.py	2006-05-11 07:17:42 +0000
+++ bzrlib/tests/workingtree_implementations/__init__.py	2006-05-18 10:31:23 +0000
@@ -55,6 +55,7 @@
         'bzrlib.tests.workingtree_implementations.test_break_lock',
         'bzrlib.tests.workingtree_implementations.test_commit',
         'bzrlib.tests.workingtree_implementations.test_is_control_filename',
+        'bzrlib.tests.workingtree_implementations.test_is_ignored',
         'bzrlib.tests.workingtree_implementations.test_pull',
         'bzrlib.tests.workingtree_implementations.test_workingtree',
         ]

=== added file 'bzrlib/tests/workingtree_implementations/test_is_ignored.py'
--- bzrlib/tests/workingtree_implementations/test_is_ignored.py	1970-01-01 00:00:00 +0000
+++ bzrlib/tests/workingtree_implementations/test_is_ignored.py	2006-05-19 00:04:00 +0000
@@ -0,0 +1,43 @@
+# (C) 2006 Canonical Ltd
+# Authors:  Robert Collins <robert.collins@canonical.com>
+#
+# This program is free software; you can redistribute it and/or modify
+# it under the terms of the GNU General Public License as published by
+# the Free Software Foundation; either version 2 of the License, or
+# (at your option) any later version.
+#
+# This program is distributed in the hope that it will be useful,
+# but WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+# GNU General Public License for more details.
+#
+# You should have received a copy of the GNU General Public License
+# along with this program; if not, write to the Free Software
+# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+
+from bzrlib.tests.workingtree_implementations import TestCaseWithWorkingTree
+
+
+class TestIsIgnored(TestCaseWithWorkingTree):
+
+    def test_is_ignored(self):
+        tree = self.make_branch_and_tree('.')
+        # this will break if a tree changes the ignored format. That is fine
+        # because at the moment tree format is orthogonal to user data, and
+        # .bzrignore is user data so must not be changed by a tree format.
+        self.build_tree_contents([
+            ('.bzrignore', './rootdir\nrandomfile*\npath/from/ro?t\n')])
+        # is_ignored returns the matching ignore regex when a path is ignored.
+        # we check some expected matches for each rule, and one or more
+        # relevant not-matches that look plausible as cases for bugs.
+        self.assertEqual('./rootdir', tree.is_ignored('rootdir'))
+        self.assertEqual(None, tree.is_ignored('foo/rootdir'))
+        self.assertEqual(None, tree.is_ignored('rootdirtrailer'))
+        self.assertEqual('randomfile*', tree.is_ignored('randomfile'))
+        self.assertEqual('randomfile*', tree.is_ignored('randomfiles'))
+        self.assertEqual('randomfile*', tree.is_ignored('foo/randomfiles'))
+        self.assertEqual(None, tree.is_ignored('randomfil'))
+        self.assertEqual(None, tree.is_ignored('foo/randomfil'))
+        self.assertEqual("path/from/ro?t", tree.is_ignored('path/from/root'))
+        self.assertEqual("path/from/ro?t", tree.is_ignored('path/from/roat'))
+        self.assertEqual(None, tree.is_ignored('roat'))

=== modified file 'bzrlib/tests/workingtree_implementations/test_workingtree.py'
--- bzrlib/tests/workingtree_implementations/test_workingtree.py	2006-05-11 07:17:42 +0000
+++ bzrlib/tests/workingtree_implementations/test_workingtree.py	2006-05-18 23:48:54 +0000
@@ -550,3 +550,14 @@
             tree.branch = tree.branch
         self.assertRaises(AttributeError, set_branch)
 
+    def test_list_files_versioned_before_ignored(self):
+        """A versioned file matching an ignore rule should not be ignored."""
+        tree = self.make_branch_and_tree('.')
+        self.build_tree(['foo.pyc'])
+        # ensure that foo.pyc is ignored
+        self.build_tree_contents([('.bzrignore', 'foo.pyc')])
+        tree.add('foo.pyc', 'anid')
+        files = sorted(list(tree.list_files()))
+        self.assertEqual((u'.bzrignore', '?', 'file', None), files[0][:-1])
+        self.assertEqual((u'foo.pyc', 'V', 'file', 'anid'), files[1][:-1])
+        self.assertEqual(2, len(files))

=== modified file 'bzrlib/transport/__init__.py'
--- bzrlib/transport/__init__.py	2006-04-12 04:45:32 +0000
+++ bzrlib/transport/__init__.py	2006-05-16 18:07:33 +0000
@@ -126,6 +126,35 @@
         urlparse.uses_netloc.append(protocol)
 
 
+def split_url(url):
+    if isinstance(url, unicode):
+        url = url.encode('utf-8')
+    (scheme, netloc, path, params,
+     query, fragment) = urlparse.urlparse(url, allow_fragments=False)
+    username = password = host = port = None
+    if '@' in netloc:
+        username, host = netloc.split('@', 1)
+        if ':' in username:
+            username, password = username.split(':', 1)
+            password = urllib.unquote(password)
+        username = urllib.unquote(username)
+    else:
+        host = netloc
+
+    if ':' in host:
+        host, port = host.rsplit(':', 1)
+        try:
+            port = int(port)
+        except ValueError:
+            # TODO: Should this be ConnectionError?
+            raise errors.TransportError('%s: invalid port number' % port)
+    host = urllib.unquote(host)
+
+    path = urllib.unquote(path)
+
+    return (scheme, username, password, host, port, path)
+
+
 class Transport(object):
     """This class encapsulates methods for retrieving or putting a file
     from/to a storage location.

=== modified file 'bzrlib/transport/ftp.py'
--- bzrlib/transport/ftp.py	2006-04-13 02:33:36 +0000
+++ bzrlib/transport/ftp.py	2006-05-22 05:32:51 +0000
@@ -25,38 +25,56 @@
 """
 
 from cStringIO import StringIO
+import asyncore
 import errno
 import ftplib
 import os
 import urllib
 import urlparse
 import stat
+import threading
 import time
 import random
 from warnings import warn
 
-
-from bzrlib.transport import Transport
-from bzrlib.errors import (TransportNotPossible, TransportError,
-                           NoSuchFile, FileExists, DirectoryNotEmpty)
+from bzrlib.transport import (
+    Transport,
+    Server,
+    split_url,
+    )
+import bzrlib.errors as errors
 from bzrlib.trace import mutter, warning
+import bzrlib.ui
+
+_have_medusa = False
+
+
+class FtpPathError(errors.PathError):
+    """FTP failed for path: %(path)s%(extra)s"""
 
 
 _FTP_cache = {}
-def _find_FTP(hostname, username, password, is_active):
+def _find_FTP(hostname, port, username, password, is_active):
     """Find an ftplib.FTP instance attached to this triplet."""
-    key = "%s|%s|%s|%s" % (hostname, username, password, is_active)
+    key = (hostname, port, username, password, is_active)
+    alt_key = (hostname, port, username, '********', is_active)
     if key not in _FTP_cache:
-        mutter("Constructing FTP instance against %r" % key)
-        _FTP_cache[key] = ftplib.FTP(hostname, username, password)
-        _FTP_cache[key].set_pasv(not is_active)
+        mutter("Constructing FTP instance against %r" % (alt_key,))
+        conn = ftplib.FTP()
+
+        conn.connect(host=hostname, port=port)
+        if username and username != 'anonymous' and not password:
+            password = bzrlib.ui.ui_factory.get_password(
+                prompt='FTP %(user)s@%(host)s password',
+                user=username, host=hostname)
+        conn.login(user=username, passwd=password)
+        conn.set_pasv(not is_active)
+
+        _FTP_cache[key] = conn
+
     return _FTP_cache[key]    
 
 
-class FtpTransportError(TransportError):
-    pass
-
-
 class FtpStatResult(object):
     def __init__(self, f, relpath):
         try:
@@ -80,36 +98,81 @@
     def __init__(self, base, _provided_instance=None):
         """Set the base path where files will be stored."""
         assert base.startswith('ftp://') or base.startswith('aftp://')
-        super(FtpTransport, self).__init__(base)
+
         self.is_active = base.startswith('aftp://')
         if self.is_active:
+            # urlparse won't handle aftp://
             base = base[1:]
-        (self._proto, self._host,
-            self._path, self._parameters,
-            self._query, self._fragment) = urlparse.urlparse(self.base)
+        if not base.endswith('/'):
+            base += '/'
+        (self._proto, self._username,
+            self._password, self._host,
+            self._port, self._path) = split_url(base)
+        base = self._unparse_url()
+
+        super(FtpTransport, self).__init__(base)
         self._FTP_instance = _provided_instance
 
+    def _unparse_url(self, path=None):
+        if path is None:
+            path = self._path
+        path = urllib.quote(path)
+        netloc = urllib.quote(self._host)
+        if self._username is not None:
+            netloc = '%s@%s' % (urllib.quote(self._username), netloc)
+        if self._port is not None:
+            netloc = '%s:%d' % (netloc, self._port)
+        return urlparse.urlunparse(('ftp', netloc, path, '', '', ''))
+
     def _get_FTP(self):
         """Return the ftplib.FTP instance for this object."""
         if self._FTP_instance is not None:
             return self._FTP_instance
         
         try:
-            username = ''
-            password = ''
-            hostname = self._host
-            if '@' in hostname:
-                username, hostname = hostname.split("@", 1)
-            if ':' in username:
-                username, password = username.split(":", 1)
-
-            self._FTP_instance = _find_FTP(hostname, username, password,
+            self._FTP_instance = _find_FTP(self._host, self._port,
+                                           self._username, self._password,
                                            self.is_active)
             return self._FTP_instance
         except ftplib.error_perm, e:
-            raise TransportError(msg="Error setting up connection: %s"
+            raise errors.TransportError(msg="Error setting up connection: %s"
                                     % str(e), orig_error=e)
 
+    def _translate_perm_error(self, err, path, extra=None, unknown_exc=FtpPathError):
+        """Try to translate an ftplib.error_perm exception.
+
+        :param err: The error to translate into a bzr error
+        :param path: The path which had problems
+        :param extra: Extra information which can be included
+        :param unknown_exc: If None, we will just raise the original exception
+                    otherwise we raise unknown_exc(path, extra=extra)
+        """
+        s = str(err).lower()
+        if not extra:
+            extra = str(err)
+        else:
+            extra += ': ' + str(err)
+        if ('no such file' in s
+            or 'could not open' in s
+            or 'no such dir' in s
+            ):
+            raise errors.NoSuchFile(path, extra=extra)
+        if ('file exists' in s):
+            raise errors.FileExists(path, extra=extra)
+        if ('not a directory' in s):
+            raise errors.PathError(path, extra=extra)
+
+        mutter('unable to understand error for path: %s: %s', path, err)
+
+        if unknown_exc:
+            raise unknown_exc(path, extra=extra)
+        # TODO: jam 20060516 Consider re-raising the error wrapped in 
+        #       something like TransportError, but this loses the traceback
+        #       Also, 'sftp' has a generic 'Failure' mode, which we use failure_exc
+        #       to handle. Consider doing something like that here.
+        #raise TransportError(msg='Error for path: %s' % (path,), orig_error=e)
+        raise
+
     def should_cache(self):
         """Return True if the data pulled across should be cached locally.
         """
@@ -127,11 +190,7 @@
     def _abspath(self, relpath):
         assert isinstance(relpath, basestring)
         relpath = urllib.unquote(relpath)
-        if isinstance(relpath, basestring):
-            relpath_parts = relpath.split('/')
-        else:
-            # TODO: Don't call this with an array - no magic interfaces
-            relpath_parts = relpath[:]
+        relpath_parts = relpath.split('/')
         if len(relpath_parts) > 1:
             if relpath_parts[0] == '':
                 raise ValueError("path %r within branch %r seems to be absolute"
@@ -153,30 +212,34 @@
         # Possibly, we could use urlparse.urljoin() here, but
         # I'm concerned about when it chooses to strip the last
         # portion of the path, and when it doesn't.
-        return '/'.join(basepath)
+        return '/'.join(basepath) or '/'
     
     def abspath(self, relpath):
         """Return the full url to the given relative path.
         This can be supplied with a string or a list
         """
         path = self._abspath(relpath)
-        return urlparse.urlunparse((self._proto,
-                self._host, path, '', '', ''))
+        return self._unparse_url(path)
 
     def has(self, relpath):
-        """Does the target location exist?
-
-        XXX: I assume we're never asked has(dirname) and thus I use
-        the FTP size command and assume that if it doesn't raise,
-        all is good.
-        """
+        """Does the target location exist?"""
+        # FIXME jam 20060516 We *do* ask about directories in the test suite
+        #       We don't seem to in the actual codebase
+        # XXX: I assume we're never asked has(dirname) and thus I use
+        # the FTP size command and assume that if it doesn't raise,
+        # all is good.
+        abspath = self._abspath(relpath)
         try:
             f = self._get_FTP()
-            s = f.size(self._abspath(relpath))
-            mutter("FTP has: %s" % self._abspath(relpath))
+            mutter('FTP has check: %s => %s', relpath, abspath)
+            s = f.size(abspath)
+            mutter("FTP has: %s", abspath)
             return True
-        except ftplib.error_perm:
-            mutter("FTP has not: %s" % self._abspath(relpath))
+        except ftplib.error_perm, e:
+            if ('is a directory' in str(e).lower()):
+                mutter("FTP has dir: %s: %s", abspath, e)
+                return True
+            mutter("FTP has not: %s: %s", abspath, e)
             return False
 
     def get(self, relpath, decode=False, retries=0):
@@ -191,26 +254,26 @@
         """
         # TODO: decode should be deprecated
         try:
-            mutter("FTP get: %s" % self._abspath(relpath))
+            mutter("FTP get: %s", self._abspath(relpath))
             f = self._get_FTP()
             ret = StringIO()
             f.retrbinary('RETR '+self._abspath(relpath), ret.write, 8192)
             ret.seek(0)
             return ret
         except ftplib.error_perm, e:
-            raise NoSuchFile(self.abspath(relpath), extra=str(e))
+            raise errors.NoSuchFile(self.abspath(relpath), extra=str(e))
         except ftplib.error_temp, e:
             if retries > _number_of_retries:
-                raise TransportError(msg="FTP temporary error during GET %s. Aborting."
+                raise errors.TransportError(msg="FTP temporary error during GET %s. Aborting."
                                      % self.abspath(relpath),
                                      orig_error=e)
             else:
-                warning("FTP temporary error: %s. Retrying." % str(e))
+                warning("FTP temporary error: %s. Retrying.", str(e))
                 self._FTP_instance = None
                 return self.get(relpath, decode, retries+1)
         except EOFError, e:
             if retries > _number_of_retries:
-                raise TransportError("FTP control connection closed during GET %s."
+                raise errors.TransportError("FTP control connection closed during GET %s."
                                      % self.abspath(relpath),
                                      orig_error=e)
             else:
@@ -229,42 +292,39 @@
 
         TODO: jam 20051215 ftp as a protocol seems to support chmod, but ftplib does not
         """
-        tmp_abspath = '%s.tmp.%.9f.%d.%d' % (self._abspath(relpath), time.time(),
+        abspath = self._abspath(relpath)
+        tmp_abspath = '%s.tmp.%.9f.%d.%d' % (abspath, time.time(),
                         os.getpid(), random.randint(0,0x7FFFFFFF))
         if not hasattr(fp, 'read'):
             fp = StringIO(fp)
         try:
-            mutter("FTP put: %s" % self._abspath(relpath))
+            mutter("FTP put: %s", abspath)
             f = self._get_FTP()
             try:
                 f.storbinary('STOR '+tmp_abspath, fp)
-                f.rename(tmp_abspath, self._abspath(relpath))
+                f.rename(tmp_abspath, abspath)
             except (ftplib.error_temp,EOFError), e:
                 warning("Failure during ftp PUT. Deleting temporary file.")
                 try:
                     f.delete(tmp_abspath)
                 except:
-                    warning("Failed to delete temporary file on the server.\nFile: %s"
-                            % tmp_abspath)
+                    warning("Failed to delete temporary file on the"
+                            " server.\nFile: %s", tmp_abspath)
                     raise e
                 raise
         except ftplib.error_perm, e:
-            if "no such file" in str(e).lower():
-                raise NoSuchFile("Error storing %s: %s"
-                                 % (self.abspath(relpath), str(e)), extra=e)
-            else:
-                raise FtpTransportError(orig_error=e)
+            self._translate_perm_error(e, abspath, extra='could not store')
         except ftplib.error_temp, e:
             if retries > _number_of_retries:
-                raise TransportError("FTP temporary error during PUT %s. Aborting."
+                raise errors.TransportError("FTP temporary error during PUT %s. Aborting."
                                      % self.abspath(relpath), orig_error=e)
             else:
-                warning("FTP temporary error: %s. Retrying." % str(e))
+                warning("FTP temporary error: %s. Retrying.", str(e))
                 self._FTP_instance = None
                 self.put(relpath, fp, mode, retries+1)
         except EOFError:
             if retries > _number_of_retries:
-                raise TransportError("FTP control connection closed during PUT %s."
+                raise errors.TransportError("FTP control connection closed during PUT %s."
                                      % self.abspath(relpath), orig_error=e)
             else:
                 warning("FTP control connection closed. Trying to reopen.")
@@ -272,70 +332,120 @@
                 self._FTP_instance = None
                 self.put(relpath, fp, mode, retries+1)
 
-
     def mkdir(self, relpath, mode=None):
         """Create a directory at the given path."""
+        abspath = self._abspath(relpath)
         try:
-            mutter("FTP mkd: %s" % self._abspath(relpath))
+            mutter("FTP mkd: %s", abspath)
             f = self._get_FTP()
-            try:
-                f.mkd(self._abspath(relpath))
-            except ftplib.error_perm, e:
-                s = str(e)
-                if 'File exists' in s:
-                    raise FileExists(self.abspath(relpath), extra=s)
-                else:
-                    raise
+            f.mkd(abspath)
         except ftplib.error_perm, e:
-            raise TransportError(orig_error=e)
+            self._translate_perm_error(e, abspath,
+                unknown_exc=errors.FileExists)
 
     def rmdir(self, rel_path):
         """Delete the directory at rel_path"""
+        abspath = self._abspath(rel_path)
         try:
-            mutter("FTP rmd: %s" % self._abspath(rel_path))
-
+            mutter("FTP rmd: %s", abspath)
             f = self._get_FTP()
-            f.rmd(self._abspath(rel_path))
+            f.rmd(abspath)
         except ftplib.error_perm, e:
-            if str(e).endswith("Directory not empty"):
-                raise DirectoryNotEmpty(self._abspath(rel_path), extra=str(e))
-            else:
-                raise TransportError(msg="Cannot remove directory at %s" % \
-                        self._abspath(rel_path), extra=str(e))
+            self._translate_perm_error(e, abspath, unknown_exc=errors.PathError)
 
-    def append(self, relpath, f):
+    def append(self, relpath, f, mode=None):
         """Append the text in the file-like object into the final
         location.
         """
-        raise TransportNotPossible('ftp does not support append()')
-
-    def copy(self, rel_from, rel_to):
-        """Copy the item at rel_from to the location at rel_to"""
-        raise TransportNotPossible('ftp does not (yet) support copy()')
+        abspath = self._abspath(relpath)
+        if self.has(relpath):
+            ftp = self._get_FTP()
+            result = ftp.size(abspath)
+        else:
+            result = 0
+
+        mutter("FTP appe to %s", abspath)
+        self._try_append(relpath, f.read(), mode)
+
+        return result
+
+    def _try_append(self, relpath, text, mode=None, retries=0):
+        """Try repeatedly to append the given text to the file at relpath.
+        
+        This is a recursive function. On errors, it will be called until the
+        number of retries is exceeded.
+        """
+        try:
+            abspath = self._abspath(relpath)
+            mutter("FTP appe (try %d) to %s", retries, abspath)
+            ftp = self._get_FTP()
+            ftp.voidcmd("TYPE I")
+            cmd = "APPE %s" % abspath
+            conn = ftp.transfercmd(cmd)
+            conn.sendall(text)
+            conn.close()
+            if mode:
+                self._setmode(relpath, mode)
+            ftp.getresp()
+        except ftplib.error_perm, e:
+            self._translate_perm_error(e, abspath, extra='error appending',
+                unknown_exc=errors.NoSuchFile)
+        except ftplib.error_temp, e:
+            if retries > _number_of_retries:
+                raise errors.TransportError("FTP temporary error during APPEND %s." \
+                        "Aborting." % abspath, orig_error=e)
+            else:
+                warning("FTP temporary error: %s. Retrying.", str(e))
+                self._FTP_instance = None
+                self._try_append(relpath, text, mode, retries+1)
+
+    def _setmode(self, relpath, mode):
+        """Set permissions on a path.
+
+        Only set permissions if the FTP server supports the 'SITE CHMOD'
+        extension.
+        """
+        try:
+            mutter("FTP site chmod: setting permissions to %s on %s",
+                str(mode), self._abspath(relpath))
+            ftp = self._get_FTP()
+            cmd = "SITE CHMOD %s %s" % (self._abspath(relpath), str(mode))
+            ftp.sendcmd(cmd)
+        except ftplib.error_perm, e:
+            # Command probably not available on this server
+            warning("FTP Could not set permissions to %s on %s. %s",
+                    str(mode), self._abspath(relpath), str(e))
+
+    # TODO: jam 20060516 I believe ftp allows you to tell an ftp server
+    #       to copy something to another machine. And you may be able
+    #       to give it its own address as the 'to' location.
+    #       So implement a fancier 'copy()'
 
     def move(self, rel_from, rel_to):
         """Move the item at rel_from to the location at rel_to"""
+        abs_from = self._abspath(rel_from)
+        abs_to = self._abspath(rel_to)
         try:
-            mutter("FTP mv: %s => %s" % (self._abspath(rel_from),
-                                         self._abspath(rel_to)))
+            mutter("FTP mv: %s => %s", abs_from, abs_to)
             f = self._get_FTP()
-            f.rename(self._abspath(rel_from), self._abspath(rel_to))
+            f.rename(abs_from, abs_to)
         except ftplib.error_perm, e:
-            raise TransportError(orig_error=e)
+            self._translate_perm_error(e, abs_from,
+                extra='unable to rename to %r' % (rel_to,), 
+                unknown_exc=errors.PathError)
 
     rename = move
 
     def delete(self, relpath):
         """Delete the item at relpath"""
+        abspath = self._abspath(relpath)
         try:
-            mutter("FTP rm: %s" % self._abspath(relpath))
+            mutter("FTP rm: %s", abspath)
             f = self._get_FTP()
-            f.delete(self._abspath(relpath))
+            f.delete(abspath)
         except ftplib.error_perm, e:
-            if str(e).endswith("No such file or directory"):
-                raise NoSuchFile(self._abspath(relpath), extra=str(e))
-            else:
-                raise TransportError(orig_error=e)
+            self._translate_perm_error(e, abspath, 'error deleting',
+                unknown_exc=errors.NoSuchFile)
 
     def listable(self):
         """See Transport.listable."""
@@ -344,16 +454,17 @@
     def list_dir(self, relpath):
         """See Transport.list_dir."""
         try:
-            mutter("FTP nlst: %s" % self._abspath(relpath))
+            mutter("FTP nlst: %s", self._abspath(relpath))
             f = self._get_FTP()
             basepath = self._abspath(relpath)
-            # FTP.nlst returns paths prefixed by relpath, strip 'em
-            the_list = f.nlst(basepath)
-            stripped = [path[len(basepath)+1:] for path in the_list]
+            paths = f.nlst(basepath)
+            # If FTP.nlst returns paths prefixed by relpath, strip 'em
+            if paths and paths[0].startswith(basepath):
+                paths = [path[len(basepath)+1:] for path in paths]
             # Remove . and .. if present, and return
-            return [path for path in stripped if path not in (".", "..")]
+            return [path for path in paths if path not in (".", "..")]
         except ftplib.error_perm, e:
-            raise TransportError(orig_error=e)
+            self._translate_perm_error(e, relpath, extra='error with list_dir')
 
     def iter_files_recursive(self):
         """See Transport.iter_files_recursive.
@@ -371,18 +482,14 @@
                 yield relpath
 
     def stat(self, relpath):
-        """Return the stat information for a file.
-        """
+        """Return the stat information for a file."""
+        abspath = self._abspath(relpath)
         try:
-            mutter("FTP stat: %s" % self._abspath(relpath))
+            mutter("FTP stat: %s", abspath)
             f = self._get_FTP()
-            return FtpStatResult(f, self._abspath(relpath))
+            return FtpStatResult(f, abspath)
         except ftplib.error_perm, e:
-            if "no such file" in str(e).lower():
-                raise NoSuchFile("Error storing %s: %s"
-                                 % (self.abspath(relpath), str(e)), extra=e)
-            else:
-                raise FtpTransportError(orig_error=e)
+            self._translate_perm_error(e, abspath, extra='error w/ stat')
 
     def lock_read(self, relpath):
         """Lock the given file for shared (read) access.
@@ -406,7 +513,207 @@
         return self.lock_read(relpath)
 
 
+class FtpServer(Server):
+    """Common code for SFTP server facilities."""
+
+    def __init__(self):
+        self._root = None
+        self._ftp_server = None
+        self._port = None
+        self._async_thread = None
+        # ftp server logs
+        self.logs = []
+
+    def get_url(self):
+        """Calculate an ftp url to this server."""
+        return 'ftp://foo:bar@localhost:%d/' % (self._port)
+
+#    def get_bogus_url(self):
+#        """Return a URL which cannot be connected to."""
+#        return 'ftp://127.0.0.1:1'
+
+    def log(self, message):
+        """This is used by medusa.ftp_server to log connections, etc."""
+        self.logs.append(message)
+
+    def setUp(self):
+
+        if not _have_medusa:
+            raise RuntimeError('Must have medusa to run the FtpServer')
+
+        self._root = os.getcwdu()
+        self._ftp_server = _ftp_server(
+            authorizer=_test_authorizer(root=self._root),
+            ip='localhost',
+            port=0, # bind to a random port
+            resolver=None,
+            logger_object=self # Use FtpServer.log() for messages
+            )
+        self._port = self._ftp_server.getsockname()[1]
+        # Don't let it loop forever, or handle an infinite number of requests.
+        # In this case it will run for 100s, or 1000 requests
+        self._async_thread = threading.Thread(target=asyncore.loop,
+                kwargs={'timeout':0.1, 'count':1000})
+        self._async_thread.setDaemon(True)
+        self._async_thread.start()
+
+    def tearDown(self):
+        """See bzrlib.transport.Server.tearDown."""
+        # have asyncore release the channel
+        self._ftp_server.del_channel()
+        asyncore.close_all()
+        self._async_thread.join()
+
+
+_ftp_channel = None
+_ftp_server = None
+_test_authorizer = None
+
+
+def _setup_medusa():
+    global _have_medusa, _ftp_channel, _ftp_server, _test_authorizer
+    try:
+        import medusa
+        import medusa.filesys
+        import medusa.ftp_server
+    except ImportError:
+        return False
+
+    _have_medusa = True
+
+    class test_authorizer(object):
+        """A custom Authorizer object for running the test suite.
+
+        The reason we cannot use dummy_authorizer, is because it sets the
+        channel to readonly, which we don't always want to do.
+        """
+
+        def __init__(self, root):
+            self.root = root
+
+        def authorize(self, channel, username, password):
+            """Return (success, reply_string, filesystem)"""
+            if not _have_medusa:
+                return 0, 'No Medusa.', None
+
+            channel.persona = -1, -1
+            if username == 'anonymous':
+                channel.read_only = 1
+            else:
+                channel.read_only = 0
+
+            return 1, 'OK.', medusa.filesys.os_filesystem(self.root)
+
+
+    class ftp_channel(medusa.ftp_server.ftp_channel):
+        """Customized ftp channel"""
+
+        def log(self, message):
+            """Redirect logging requests."""
+            mutter('_ftp_channel: %s', message)
+            
+        def log_info(self, message, type='info'):
+            """Redirect logging requests."""
+            mutter('_ftp_channel %s: %s', type, message)
+            
+        def cmd_rnfr(self, line):
+            """Prepare for renaming a file."""
+            self._renaming = line[1]
+            self.respond('350 Ready for RNTO')
+            # TODO: jam 20060516 in testing, the ftp server seems to
+            #       check that the file already exists, or it sends
+            #       550 RNFR command failed
+
+        def cmd_rnto(self, line):
+            """Rename a file based on the target given.
+
+            rnto must be called after calling rnfr.
+            """
+            if not self._renaming:
+                self.respond('503 RNFR required first.')
+            pfrom = self.filesystem.translate(self._renaming)
+            self._renaming = None
+            pto = self.filesystem.translate(line[1])
+            try:
+                os.rename(pfrom, pto)
+            except (IOError, OSError), e:
+                # TODO: jam 20060516 return custom responses based on
+                #       why the command failed
+                self.respond('550 RNTO failed: %s' % (e,))
+            except:
+                self.respond('550 RNTO failed')
+                # For a test server, we will go ahead and just die
+                raise
+            else:
+                self.respond('250 Rename successful.')
+
+        def cmd_size(self, line):
+            """Return the size of a file
+
+            This is overloaded to help the test suite determine if the 
+            target is a directory.
+            """
+            filename = line[1]
+            if not self.filesystem.isfile(filename):
+                if self.filesystem.isdir(filename):
+                    self.respond('550 "%s" is a directory' % (filename,))
+                else:
+                    self.respond('550 "%s" is not a file' % (filename,))
+            else:
+                self.respond('213 %d' 
+                    % (self.filesystem.stat(filename)[stat.ST_SIZE]),)
+
+        def cmd_mkd(self, line):
+            """Create a directory.
+
+            Overloaded because default implementation does not distinguish
+            *why* it cannot make a directory.
+            """
+            if len (line) != 2:
+                self.command_not_understood (string.join (line))
+            else:
+                path = line[1]
+                try:
+                    self.filesystem.mkdir (path)
+                    self.respond ('257 MKD command successful.')
+                except (IOError, OSError), e:
+                    self.respond ('550 error creating directory: %s' % (e,))
+                except:
+                    self.respond ('550 error creating directory.')
+
+
+    class ftp_server(medusa.ftp_server.ftp_server):
+        """Customize the behavior of the Medusa ftp_server.
+
+        There are a few warts on the ftp_server, based on how it expects
+        to be used.
+        """
+        _renaming = None
+        ftp_channel_class = ftp_channel
+
+        def __init__(self, *args, **kwargs):
+            mutter('Initializing _ftp_server: %r, %r', args, kwargs)
+            medusa.ftp_server.ftp_server.__init__(self, *args, **kwargs)
+
+        def log(self, message):
+            """Redirect logging requests."""
+            mutter('_ftp_server: %s', message)
+
+        def log_info(self, message, type='info'):
+            """Override the asyncore.log_info so we don't stipple the screen."""
+            mutter('_ftp_server %s: %s', type, message)
+
+    _test_authorizer = test_authorizer
+    _ftp_channel = ftp_channel
+    _ftp_server = ftp_server
+
+    return True
+
+
 def get_test_permutations():
     """Return the permutations to be used in testing."""
-    warn("There are no FTP transport provider tests yet.")
-    return []
+    if not _setup_medusa():
+        warn("You must install medusa (http://www.amk.ca/python/code/medusa.html) for FTP tests")
+        return []
+    else:
+        return [(FtpTransport, FtpServer)]

=== modified file 'bzrlib/transport/sftp.py'
--- bzrlib/transport/sftp.py	2006-05-18 03:47:41 +0000
+++ bzrlib/transport/sftp.py	2006-05-19 00:35:14 +0000
@@ -44,6 +44,7 @@
 from bzrlib.transport import (
     register_urlparse_netloc_protocol,
     Server,
+    split_url,
     Transport,
     urlescape,
     )
@@ -662,31 +663,8 @@
         return urlparse.urlunparse(('sftp', netloc, path, '', '', ''))
 
     def _split_url(self, url):
-        if isinstance(url, unicode):
-            url = url.encode('utf-8')
-        (scheme, netloc, path, params,
-         query, fragment) = urlparse.urlparse(url, allow_fragments=False)
+        (scheme, username, password, host, port, path) = split_url(url)
         assert scheme == 'sftp'
-        username = password = host = port = None
-        if '@' in netloc:
-            username, host = netloc.split('@', 1)
-            if ':' in username:
-                username, password = username.split(':', 1)
-                password = urllib.unquote(password)
-            username = urllib.unquote(username)
-        else:
-            host = netloc
-
-        if ':' in host:
-            host, port = host.rsplit(':', 1)
-            try:
-                port = int(port)
-            except ValueError:
-                # TODO: Should this be ConnectionError?
-                raise TransportError('%s: invalid port number' % port)
-        host = urllib.unquote(host)
-
-        path = urllib.unquote(path)
 
         # the initial slash should be removed from the path, and treated
         # as a homedir relative path (the path begins with a double slash

=== modified file 'bzrlib/workingtree.py'
--- bzrlib/workingtree.py	2006-05-17 03:40:51 +0000
+++ bzrlib/workingtree.py	2006-05-19 07:57:09 +0000
@@ -39,13 +39,15 @@
 # At the moment they may alias the inventory and have old copies of it in
 # memory.  (Now done? -- mbp 20060309)
 
+from binascii import hexlify
 from copy import deepcopy
 from cStringIO import StringIO
 import errno
 import fnmatch
 import os
+import re
 import stat
- 
+from time import time
 
 from bzrlib.atomicfile import AtomicFile
 from bzrlib.branch import (Branch,
@@ -80,7 +82,7 @@
                             pumpfile,
                             safe_unicode,
                             splitpath,
-                            rand_bytes,
+                            rand_chars,
                             normpath,
                             realpath,
                             relpath,
@@ -101,33 +103,43 @@
 import bzrlib.xml5
 
 
+# the regex here does the following:
+# 1) remove any weird characters; we don't escape them but rather
+# just pull them out
+ # 2) match leading '.'s to make it not hidden
+_gen_file_id_re = re.compile(r'[^\w.]|(^\.*)')
+_gen_id_suffix = None
+_gen_id_serial = 0
+
+
+def _next_id_suffix():
+    """Create a new file id suffix that is reasonably unique.
+    
+    On the first call we combine the current time with 64 bits of randomness
+    to give a highly probably globally unique number. Then each call in the same
+    process adds 1 to a serial number we append to that unique value.
+    """
+    # XXX TODO: change bzrlib.add.smart_add to call workingtree.add() rather 
+    # than having to move the id randomness out of the inner loop like this.
+    # XXX TODO: for the global randomness this uses we should add the thread-id
+    # before the serial #.
+    global _gen_id_suffix, _gen_id_serial
+    if _gen_id_suffix is None:
+        _gen_id_suffix = "-%s-%s-" % (compact_date(time()), rand_chars(16))
+    _gen_id_serial += 1
+    return _gen_id_suffix + str(_gen_id_serial)
+
+
 def gen_file_id(name):
-    """Return new file id.
-
-    This should probably generate proper UUIDs, but for the moment we
-    cope with just randomness because running uuidgen every time is
-    slow."""
-    import re
-    from binascii import hexlify
-    from time import time
-
-    # get last component
-    idx = name.rfind('/')
-    if idx != -1:
-        name = name[idx+1 : ]
-    idx = name.rfind('\\')
-    if idx != -1:
-        name = name[idx+1 : ]
-
-    # make it not a hidden file
-    name = name.lstrip('.')
-
-    # remove any wierd characters; we don't escape them but rather
-    # just pull them out
-    name = re.sub(r'[^\w.]', '', name)
-
-    s = hexlify(rand_bytes(8))
-    return '-'.join((name, compact_date(time()), s))
+    """Return new file id for the basename 'name'.
+
+    The uniqueness is supplied from _next_id_suffix.
+    """
+    # XXX TODO: squash the filename to lowercase.
+    # XXX TODO: truncate the filename to something like 20 or 30 chars.
+    # XXX TODO: consider what to do with ids that look like illegal filepaths
+    # on platforms we support.
+    return _gen_file_id_re.sub('', name) + _next_id_suffix()
 
 
 def gen_root_id():
@@ -308,15 +320,14 @@
     def is_control_filename(self, filename):
         """True if filename is the name of a control file in this tree.
         
+        :param filename: A filename within the tree. This is a relative path
+        from the root of this tree.
+
         This is true IF and ONLY IF the filename is part of the meta data
         that bzr controls in this tree. I.E. a random .bzr directory placed
         on disk will not be a control file for this tree.
         """
-        try:
-            self.bzrdir.transport.relpath(self.abspath(filename))
-            return True
-        except errors.PathNotChild:
-            return False
+        return self.bzrdir.is_control_filename(filename)
 
     @staticmethod
     def open(path=None, _unsupported=False):
@@ -423,9 +434,13 @@
         """
         return bzrdir.BzrDir.create_standalone_workingtree(directory)
 
-    def relpath(self, abs):
-        """Return the local path portion from a given absolute path."""
-        return relpath(self.basedir, abs)
+    def relpath(self, path):
+        """Return the local path portion from a given path.
+        
+        The path may be absolute or relative. If its a relative path it is 
+        interpreted relative to the python current working directory.
+        """
+        return relpath(self.basedir, path)
 
     def has_filename(self, filename):
         return bzrlib.osutils.lexists(self.abspath(filename))
@@ -588,10 +603,10 @@
                                'i.e. regular file, symlink or directory): %s' % quotefn(f))
 
             if file_id is None:
-                file_id = gen_file_id(f)
-            inv.add_path(f, kind=kind, file_id=file_id)
+                inv.add_path(f, kind=kind)
+            else:
+                inv.add_path(f, kind=kind, file_id=file_id)
 
-            mutter("add file %s file_id:{%s} kind=%r" % (f, file_id, kind))
         self._write_inventory(inv)
 
     @needs_write_lock
@@ -966,6 +981,60 @@
                 subp = appendpath(path, subf)
                 yield subp
 
+    def _translate_ignore_rule(self, rule):
+        """Translate a single ignore rule to a regex.
+
+        There are two types of ignore rules.  Those that do not contain a / are
+        matched against the tail of the filename (that is, they do not care
+        what directory the file is in.)  Rules which do contain a slash must
+        match the entire path.  As a special case, './' at the start of the
+        string counts as a slash in the string but is removed before matching
+        (e.g. ./foo.c, ./src/foo.c)
+
+        :return: The translated regex.
+        """
+        if rule[:2] in ('./', '.\\'):
+            # rootdir rule
+            result = fnmatch.translate(rule[2:])
+        elif '/' in rule or '\\' in rule:
+            # path prefix 
+            result = fnmatch.translate(rule)
+        else:
+            # default rule style.
+            result = "(?:.*/)?(?!.*/)" + fnmatch.translate(rule)
+        assert result[-1] == '$', "fnmatch.translate did not add the expected $"
+        return "(" + result + ")"
+
+    def _combine_ignore_rules(self, rules):
+        """Combine a list of ignore rules into a single regex object.
+
+        Each individual rule is combined with | to form a big regex, which then
+        has $ added to it to form something like ()|()|()$. The group index for
+        each subregex's outermost group is placed in a dictionary mapping back 
+        to the rule. This allows quick identification of the matching rule that
+        triggered a match.
+        :return: a list of the compiled regex and the matching-group index 
+        dictionaries. We return a list because python complains if you try to 
+        combine more than 100 regexes.
+        """
+        result = []
+        groups = {}
+        next_group = 0
+        translated_rules = []
+        for rule in rules:
+            translated_rule = self._translate_ignore_rule(rule)
+            compiled_rule = re.compile(translated_rule)
+            groups[next_group] = rule
+            next_group += compiled_rule.groups
+            translated_rules.append(translated_rule)
+            if next_group == 99:
+                result.append((re.compile("|".join(translated_rules)), groups))
+                groups = {}
+                next_group = 0
+                translated_rules = []
+        if len(translated_rules):
+            result.append((re.compile("|".join(translated_rules)), groups))
+        return result
 
     def ignored_files(self):
         """Yield list of PATH, IGNORE_PATTERN"""
@@ -974,7 +1043,6 @@
             if pat != None:
                 yield subp, pat
 
-
     def get_ignore_list(self):
         """Return list of ignore patterns.
 
@@ -988,8 +1056,18 @@
             f = self.get_file_byname(bzrlib.IGNORE_FILENAME)
             l.extend([line.rstrip("\n\r") for line in f.readlines()])
         self._ignorelist = l
+        self._ignore_regex = self._combine_ignore_rules(l)
         return l
 
+    def _get_ignore_rules_as_regex(self):
+        """Return a regex of the ignore rules and a mapping dict.
+
+        :return: (ignore rules compiled regex, dictionary mapping rule group 
+        indices to original rule.)
+        """
+        if getattr(self, '_ignorelist', None) is None:
+            self.get_ignore_list()
+        return self._ignore_regex
 
     def is_ignored(self, filename):
         r"""Check whether the filename matches an ignore pattern.
@@ -1009,23 +1087,17 @@
         # treat dotfiles correctly and allows * to match /.
         # Eventually it should be replaced with something more
         # accurate.
-        
-        basename = splitpath(filename)[-1]
-        for pat in self.get_ignore_list():
-            if '/' in pat or '\\' in pat:
-                
-                # as a special case, you can put ./ at the start of a
-                # pattern; this is good to match in the top-level
-                # only;
-                if pat[:2] in ('./', '.\\'):
-                    newpat = pat[2:]
-                else:
-                    newpat = pat
-                if fnmatch.fnmatchcase(filename, newpat):
-                    return pat
-            else:
-                if fnmatch.fnmatchcase(basename, pat):
-                    return pat
+    
+        rules = self._get_ignore_rules_as_regex()
+        for regex, mapping in rules:
+            match = regex.match(filename)
+            if match is not None:
+                # one or more of the groups in mapping will have a non-None group 
+                # match.
+                groups = match.groups()
+                rules = [mapping[group] for group in 
+                    mapping if groups[group] is not None]
+                return rules[0]
         return None
 
     def kind(self, file_id):

=== modified file 'bzrlib/xml_serializer.py'
--- bzrlib/xml_serializer.py	2006-01-19 21:25:01 +0000
+++ bzrlib/xml_serializer.py	2006-05-23 00:03:45 +0000
@@ -71,3 +71,71 @@
 
     def _read_element(self, f):
         return ElementTree().parse(f)
+
+
+# performance tuning for elementree's serialiser. THis should be
+# sent upstream - RBC 20060523.
+# the functions here are patched into elementree at runtime.
+import elementtree.ElementTree
+import re
+escape_re = re.compile("[&'\"<>]")
+escape_map = {
+    "&":'&amp;',
+    "'":"&apos;", # FIXME: overkill
+    "\"":"&quot;",
+    "<":"&lt;",
+    ">":"&gt;",
+    }
+def _escape_replace(match, map=escape_map):
+    return map[match.group()]
+ 
+def _escape_attrib(text, encoding=None, replace=None):
+    # escape attribute value
+    try:
+        if encoding:
+            try:
+                text = elementtree.ElementTree._encode(text, encoding)
+            except UnicodeError:
+                return elementtree.ElementTree._encode_entity(text)
+        if replace is None:
+            return escape_re.sub(_escape_replace, text)
+        else:
+            text = replace(text, "&", "&amp;")
+            text = replace(text, "'", "&apos;") # FIXME: overkill
+            text = replace(text, "\"", "&quot;")
+            text = replace(text, "<", "&lt;")
+            text = replace(text, ">", "&gt;")
+            return text
+    except (TypeError, AttributeError):
+        elementtree.ElementTree._raise_serialization_error(text)
+
+elementtree.ElementTree._escape_attrib = _escape_attrib
+
+escape_cdata_re = re.compile("[&<>]")
+escape_cdata_map = {
+    "&":'&amp;',
+    "<":"&lt;",
+    ">":"&gt;",
+    }
+def _escape_cdata_replace(match, map=escape_cdata_map):
+    return map[match.group()]
+ 
+def _escape_cdata(text, encoding=None, replace=None):
+    # escape character data
+    try:
+        if encoding:
+            try:
+                text = elementtree.ElementTree._encode(text, encoding)
+            except UnicodeError:
+                return elementtree.ElementTree._encode_entity(text)
+        if replace is None:
+            return escape_cdata_re.sub(_escape_cdata_replace, text)
+        else:
+            text = replace(text, "&", "&amp;")
+            text = replace(text, "<", "&lt;")
+            text = replace(text, ">", "&gt;")
+            return text
+    except (TypeError, AttributeError):
+        elementtree.ElementTree._raise_serialization_error(text)
+
+elementtree.ElementTree._escape_cdata = _escape_cdata

